class PresenceStateContainer:
    channel_states: dict

    def __init__(self):
        self.channel_states = {}

    def register_state(self, state: dict, channels: list):
        for channel in channels:
            self.channel_states[channel] = state

    def get_state(self, channels: list):
        return {channel: self.channel_states[channel] for channel in channels if channel in self.channel_states}

    def get_channels_states(self, channels: list):
        return {channel: self.channel_states[channel] for channel in channels if channel in self.channel_states}



import hashlib
import json
import random
import logging


from base64 import decodebytes, encodebytes, b64decode, b64encode
from Cryptodome.Cipher import AES
from Cryptodome.Util.Padding import pad, unpad
from pubnub.crypto_core import PubNubCrypto, PubNubCryptor, PubNubLegacyCryptor, PubNubAesCbcCryptor, CryptoHeader, \
    CryptorPayload
from pubnub.exceptions import PubNubException
from typing import Union, Dict


Initial16bytes = '0123456789012345'


class PubNubCryptodome(PubNubCrypto):
    mode = AES.MODE_CBC
    fallback_mode = None

    def __init__(self, pubnub_config):
        super().__init__(pubnub_config)
        self.mode = pubnub_config.cipher_mode
        self.fallback_mode = pubnub_config.fallback_cipher_mode

    def encrypt(self, key, msg, use_random_iv=False):
        secret = self.get_secret(key)
        initialization_vector = self.get_initialization_vector(use_random_iv)

        cipher = AES.new(bytes(secret[0:32], 'utf-8'), self.mode, bytes(initialization_vector, 'utf-8'))
        encrypted_message = cipher.encrypt(self.pad(msg.encode('utf-8')))
        msg_with_iv = self.append_random_iv(encrypted_message, use_random_iv, bytes(initialization_vector, "utf-8"))

        return encodebytes(msg_with_iv).decode('utf-8').replace("\n", "")

    def decrypt(self, key, msg, use_random_iv=False):
        secret = self.get_secret(key)

        decoded_message = decodebytes(msg.encode("utf-8"))
        initialization_vector, extracted_message = self.extract_random_iv(decoded_message, use_random_iv)
        cipher = AES.new(bytes(secret[0:32], "utf-8"), self.mode, initialization_vector)
        try:
            plain = self.depad((cipher.decrypt(extracted_message)).decode('utf-8'))
        except UnicodeDecodeError as e:
            if not self.fallback_mode:
                raise e

            cipher = AES.new(bytes(secret[0:32], "utf-8"), self.fallback_mode, initialization_vector)
            plain = self.depad((cipher.decrypt(extracted_message)).decode('utf-8'))

        try:
            return json.loads(plain)
        except Exception:
            return plain

    def append_random_iv(self, message, use_random_iv, initialization_vector):
        if self.pubnub_configuration.use_random_initialization_vector or use_random_iv:
            return initialization_vector + message
        else:
            return message

    def extract_random_iv(self, message, use_random_iv):
        if self.pubnub_configuration.use_random_initialization_vector or use_random_iv:
            return message[0:16], message[16:]
        else:
            return bytes(Initial16bytes, "utf-8"), message

    def get_initialization_vector(self, use_random_iv):
        if self.pubnub_configuration.use_random_initialization_vector or use_random_iv:
            return "{0:016}".format(random.randint(0, 9999999999999999))
        else:
            return Initial16bytes

    def pad(self, msg, block_size=16):
        padding = block_size - (len(msg) % block_size)
        return msg + (chr(padding) * padding).encode('utf-8')

    def depad(self, msg):
        return msg[0:-ord(msg[-1])]

    def get_secret(self, key):
        return hashlib.sha256(key.encode("utf-8")).hexdigest()


class PubNubFileCrypto(PubNubCryptodome):
    def encrypt(self, key, file, use_random_iv=True):

        secret = self.get_secret(key)
        initialization_vector = self.get_initialization_vector(use_random_iv)
        cipher = AES.new(bytes(secret[0:32], "utf-8"), self.mode, bytes(initialization_vector, 'utf-8'))
        initialization_vector = bytes(initialization_vector, 'utf-8')

        return self.append_random_iv(
            cipher.encrypt(pad(file, 16)),
            use_random_iv=True,
            initialization_vector=initialization_vector
        )

    def decrypt(self, key, file, use_random_iv=True):
        secret = self.get_secret(key)
        initialization_vector, extracted_file = self.extract_random_iv(file, use_random_iv)
        try:
            cipher = AES.new(bytes(secret[0:32], "utf-8"), self.mode, initialization_vector)
            result = unpad(cipher.decrypt(extracted_file), 16)
        except ValueError:
            cipher = AES.new(bytes(secret[0:32], "utf-8"), self.fallback_mode, initialization_vector)
            result = unpad(cipher.decrypt(extracted_file), 16)

        return result


class PubNubCryptoModule(PubNubCrypto):
    FALLBACK_CRYPTOR_ID: str = '0000'
    cryptor_map = {}
    default_cryptor_id: str

    def __init__(self, cryptor_map: Dict[str, PubNubCryptor], default_cryptor: PubNubCryptor):
        self.cryptor_map = cryptor_map
        self.default_cryptor_id = default_cryptor.CRYPTOR_ID

    def _validate_cryptor_id(self, cryptor_id: str) -> str:
        cryptor_id = cryptor_id or self.default_cryptor_id

        if len(cryptor_id) != 4:
            logging.error(f'Malformed cryptor id: {cryptor_id}')
            raise PubNubException('Malformed cryptor id')

        if cryptor_id not in self.cryptor_map.keys():
            logging.error(f'Unsupported cryptor: {cryptor_id}')
            raise PubNubException('unknown cryptor error')
        return cryptor_id

    def _get_cryptor(self, cryptor_id):
        if not cryptor_id or cryptor_id not in self.cryptor_map:
            raise PubNubException('unknown cryptor error')
        return self.cryptor_map[cryptor_id]

    # encrypt string
    def encrypt(self, message: str, cryptor_id: str = None) -> str:
        if not len(message):
            raise PubNubException('encryption error')
        cryptor_id = self._validate_cryptor_id(cryptor_id)
        data = message.encode('utf-8')
        crypto_payload = self.cryptor_map[cryptor_id].encrypt(data)
        header = self.encode_header(cryptor_id=cryptor_id, cryptor_data=crypto_payload['cryptor_data'])
        return b64encode(header + crypto_payload['data']).decode()

    def decrypt(self, message):
        data = b64decode(message)
        header = self.decode_header(data)
        if header:
            cryptor_id = header['cryptor_id']
            payload = CryptorPayload(data=data[header['length']:], cryptor_data=header['cryptor_data'])
        if not header:
            cryptor_id = self.FALLBACK_CRYPTOR_ID
            payload = CryptorPayload(data=data)

        if not len(payload['data']):
            raise PubNubException('decryption error')

        if cryptor_id not in self.cryptor_map.keys():
            raise PubNubException('unknown cryptor error')

        message = self._get_cryptor(cryptor_id).decrypt(payload)
        try:
            return json.loads(message)
        except Exception:
            return message

    def encrypt_file(self, file_data, cryptor_id: str = None):
        if not len(file_data):
            raise PubNubException('encryption error')
        cryptor_id = self._validate_cryptor_id(cryptor_id)
        crypto_payload = self.cryptor_map[cryptor_id].encrypt(file_data)
        header = self.encode_header(cryptor_id=cryptor_id, cryptor_data=crypto_payload['cryptor_data'])
        return header + crypto_payload['data']

    def decrypt_file(self, file_data):
        header = self.decode_header(file_data)
        if header:
            cryptor_id = header['cryptor_id']
            payload = CryptorPayload(data=file_data[header['length']:], cryptor_data=header['cryptor_data'])
        else:
            cryptor_id = self.FALLBACK_CRYPTOR_ID
            payload = CryptorPayload(data=file_data)

        if not len(payload['data']):
            raise PubNubException('decryption error')

        if cryptor_id not in self.cryptor_map.keys():
            raise PubNubException('unknown cryptor error')

        return self._get_cryptor(cryptor_id).decrypt(payload, binary_mode=True)

    def encode_header(self, cryptor_id: str = None, cryptor_data: any = None) -> str:
        if cryptor_id == self.FALLBACK_CRYPTOR_ID:
            return b''
        if cryptor_data and len(cryptor_data) > 65535:
            raise PubNubException('Cryptor data is too long')
        cryptor_id = self._validate_cryptor_id(cryptor_id)

        sentinel = b'PNED'
        version = CryptoHeader.header_ver.to_bytes(1, byteorder='big')
        crid = bytes(cryptor_id, 'utf-8')

        if cryptor_data:
            crd = cryptor_data
            cryptor_data_len = len(cryptor_data)
        else:
            crd = b''
            cryptor_data_len = 0

        if cryptor_data_len < 255:
            crlen = cryptor_data_len.to_bytes(1, byteorder='big')
        else:
            crlen = b'\xff' + cryptor_data_len.to_bytes(2, byteorder='big')
        return sentinel + version + crid + crlen + crd

    def decode_header(self, header: bytes) -> Union[None, CryptoHeader]:
        try:
            sentinel = header[:4]
            if sentinel != b'PNED':
                return False
        except ValueError:
            return False

        try:
            header_version = header[4]
            if header_version > CryptoHeader.header_ver:
                raise PubNubException('unknown cryptor error')

            cryptor_id = header[5:9].decode()
            crlen = header[9]
            if crlen < 255:
                cryptor_data = header[10: 10 + crlen]
                hlen = 10 + crlen
            else:
                crlen = int(header[10:12].hex(), 16)
                cryptor_data = header[12:12 + crlen]
                hlen = 12 + crlen

            return CryptoHeader(sentinel=sentinel, header_ver=header_version, cryptor_id=cryptor_id,
                                cryptor_data=cryptor_data, length=hlen)
        except IndexError:
            raise PubNubException('decryption error')


class LegacyCryptoModule(PubNubCryptoModule):
    def __init__(self, config) -> None:
        cryptor_map = {
            PubNubLegacyCryptor.CRYPTOR_ID: PubNubLegacyCryptor(config.cipher_key,
                                                                config.use_random_initialization_vector,
                                                                config.cipher_mode,
                                                                config.fallback_cipher_mode),
            PubNubAesCbcCryptor.CRYPTOR_ID: PubNubAesCbcCryptor(config.cipher_key),
        }
        super().__init__(cryptor_map, PubNubLegacyCryptor)


class AesCbcCryptoModule(PubNubCryptoModule):
    def __init__(self, config) -> None:
        cryptor_map = {
            PubNubLegacyCryptor.CRYPTOR_ID: PubNubLegacyCryptor(config.cipher_key,
                                                                config.use_random_initialization_vector,
                                                                config.cipher_mode,
                                                                config.fallback_cipher_mode),
            PubNubAesCbcCryptor.CRYPTOR_ID: PubNubAesCbcCryptor(config.cipher_key),
        }
        super().__init__(cryptor_map, PubNubAesCbcCryptor)



import logging

from typing import List, Optional

from pubnub.event_engine.models import events, invocations, states
from pubnub.event_engine.dispatcher import Dispatcher


class StateMachine:
    _current_state: states.PNState
    _context: states.PNContext
    _invocations: List[invocations.PNInvocation]
    _enabled: bool

    def __init__(self, initial_state: states.PNState, dispatcher_class: Optional[Dispatcher] = None,
                 name: Optional[str] = None) -> None:
        self._context = states.PNContext()
        self._current_state = initial_state(self._context)
        self._listeners = {}
        self._invocations = []
        if dispatcher_class is None:
            dispatcher_class = Dispatcher
        self._dispatcher = dispatcher_class(self)
        self._enabled = True
        self._name = name
        self.logger = logging.getLogger("pubnub" if not name else f"pubnub.{name}")

    def __del__(self):
        self.logger.debug('Shutting down StateMachine')
        self._enabled = False

    def get_state_name(self):
        return self._current_state.__class__.__name__

    def get_context(self) -> states.PNContext:
        return self._current_state._context

    def get_dispatcher(self) -> Dispatcher:
        return self._dispatcher

    def trigger(self, event: events.PNEvent) -> states.PNTransition:
        self.logger.debug(f'Current State: {self.get_state_name()}')
        self.logger.debug(f'Triggered event: {event.__class__.__name__}({event.__dict__}) on {self.get_state_name()}')

        if not self._enabled:
            self.logger.error('EventEngine is not enabled')
            return False

        if event.get_name() in self._current_state._transitions:
            self._invocations.clear()
            invocation = self._current_state.on_exit()

            if invocation:
                self.logger.debug(f'Invoke effect: {invocation.__class__.__name__}')
                self._invocations.append(invocation)

            transition: states.PNTransition = self._current_state.on(event, self._context)
            self._current_state = transition.state(self._current_state.get_context())
            self._context = transition.context

            if transition.invocation:
                if isinstance(transition.invocation, list):
                    self.logger.debug('unpacking list')
                    for invocation in transition.invocation:
                        self.logger.debug(f'Invoke effect: {invocation.__class__.__name__}')
                        self._invocations.append(invocation)
                else:
                    self.logger.debug(f'Invoke effect: {transition.invocation.__class__.__name__}')
                    self._invocations.append(transition.invocation)

            invocation = self._current_state.on_enter(self._context)

            if invocation:
                self.logger.debug(f'Invoke effect: {invocation.__class__.__name__}')
                self._invocations.append(invocation)

        else:
            self.logger.warning(f'Unhandled event: {event.get_name()} in {self.get_state_name()}')

        self.dispatch_effects()

    def dispatch_effects(self):
        for invocation in self._invocations:
            self.logger.debug(f'Dispatching {invocation.__class__.__name__} {id(invocation)}')
            self._dispatcher.dispatch_effect(invocation)

        self._invocations.clear()

    def stop(self):
        self._enabled = False

    @property
    def name(self):
        return self._name



from typing import List, Union
from pubnub.exceptions import PubNubException
from pubnub.enums import PNStatusCategory


class PNInvocation:
    pass


class PNManageableInvocation(PNInvocation):
    pass


class PNCancelInvocation(PNInvocation):
    cancel_effect: str


class HandshakeInvocation(PNManageableInvocation):
    def __init__(self, channels: Union[None, List[str]] = None, groups: Union[None, List[str]] = None,
                 timetoken: Union[None, int] = None) -> None:
        super().__init__()
        self.channels = channels
        self.groups = groups
        self.timetoken = timetoken


class CancelHandshakeInvocation(PNCancelInvocation):
    cancel_effect = HandshakeInvocation.__name__


class ReceiveMessagesInvocation(PNManageableInvocation):
    def __init__(self,
                 channels: Union[None, List[str]] = None,
                 groups: Union[None, List[str]] = None,
                 timetoken: Union[None, str] = None,
                 region: Union[None, int] = None
                 ) -> None:
        super().__init__()
        self.channels = channels
        self.groups = groups
        self.timetoken = timetoken
        self.region = region


class CancelReceiveMessagesInvocation(PNCancelInvocation):
    cancel_effect = ReceiveMessagesInvocation.__name__


class ReconnectInvocation(PNManageableInvocation):
    def __init__(self,
                 channels: Union[None, List[str]] = None,
                 groups: Union[None, List[str]] = None,
                 timetoken: Union[None, str] = None,
                 region: Union[None, int] = None,
                 attempts: Union[None, int] = None,
                 reason: Union[None, PubNubException] = None
                 ) -> None:
        self.channels = channels
        self.groups = groups
        self.attempts = attempts
        self.reason = reason
        self.timetoken = timetoken
        self.region = region


class HandshakeReconnectInvocation(ReconnectInvocation):
    pass


class CancelHandshakeReconnectInvocation(PNCancelInvocation):
    cancel_effect = HandshakeReconnectInvocation.__name__


class ReceiveReconnectInvocation(ReconnectInvocation):
    pass


class CancelReceiveReconnectInvocation(PNCancelInvocation):
    cancel_effect = ReceiveReconnectInvocation.__name__


class PNEmittableInvocation(PNInvocation):
    pass


class EmitMessagesInvocation(PNEmittableInvocation):
    def __init__(self, messages: Union[None, List[str]]) -> None:
        super().__init__()
        self.messages = messages


class EmitStatusInvocation(PNEmittableInvocation):
    def __init__(self, status: Union[None, PNStatusCategory]) -> None:
        super().__init__()
        self.status = status


"""
    Presence Effects
"""


class HeartbeatInvocation(PNManageableInvocation):
    def __init__(self, channels: Union[None, List[str]] = None, groups: Union[None, List[str]] = None) -> None:
        super().__init__()
        self.channels = channels
        self.groups = groups


class HeartbeatWaitInvocation(PNManageableInvocation):
    def __init__(self, time) -> None:
        self.wait_time = time
        super().__init__()


class HeartbeatCancelWaitInvocation(PNCancelInvocation):
    cancel_effect = HeartbeatWaitInvocation.__name__


class HeartbeatLeaveInvocation(PNManageableInvocation):
    def __init__(self, channels: Union[None, List[str]] = None, groups: Union[None, List[str]] = None,
                 suppress_leave: bool = False) -> None:
        super().__init__()
        self.channels = channels
        self.groups = groups
        self.suppress_leave = suppress_leave


class HeartbeatDelayedHeartbeatInvocation(PNManageableInvocation):
    def __init__(self,
                 channels: Union[None, List[str]] = None,
                 groups: Union[None, List[str]] = None,
                 attempts: Union[None, int] = None,
                 reason: Union[None, PubNubException] = None):
        super().__init__()
        self.channels = channels
        self.groups = groups
        self.attempts = attempts
        self.reason = reason


class HeartbeatCancelDelayedHeartbeatInvocation(PNCancelInvocation):
    cancel_effect = HeartbeatDelayedHeartbeatInvocation.__name__



class SubscribeOperation(object):
    def __init__(self, channels=None, channel_groups=None, presence_enabled=None, timetoken=None):
        assert isinstance(channels, (list, tuple))
        assert isinstance(channel_groups, (list, tuple))
        assert isinstance(presence_enabled, bool)
        assert isinstance(timetoken, int)

        self.channels = channels
        self.channel_groups = channel_groups
        self.presence_enabled = presence_enabled
        self.timetoken = timetoken

    @property
    def channels_with_pressence(self):
        if not self.presence_enabled:
            return self.channels
        return self.channels + [ch + '-pnpres' for ch in self.channels]

    @property
    def groups_with_pressence(self):
        if not self.presence_enabled:
            return self.channel_groups
        return self.channel_groups + [ch + '-pnpres' for ch in self.channel_groups]


class UnsubscribeOperation(object):
    def __init__(self, channels=None, channel_groups=None):
        assert isinstance(channels, (list, tuple))
        assert isinstance(channel_groups, (list, tuple))

        self.channels = channels
        self.channel_groups = channel_groups

    def get_subscribed_channels(self, channels, with_presence=False) -> list:
        result = [ch for ch in channels if ch not in self.channels and not ch.endswith('-pnpres')]
        if not with_presence:
            return result
        return result + [ch + '-pnpres' for ch in result]

    def get_subscribed_channel_groups(self, channel_groups, with_presence=False) -> list:
        result = [grp for grp in channel_groups if grp not in self.channel_groups and not grp.endswith('-pnpres')]
        if not with_presence:
            return result
        return result + [grp + '-pnpres' for grp in result]


class StateOperation(object):
    def __init__(self, channels=None, channel_groups=None, state=None):
        self.channels = channels
        self.channel_groups = channel_groups
        self.state = state



import hashlib
import json
import random
import secrets

from abc import abstractmethod
from Cryptodome.Cipher import AES
from Cryptodome.Util.Padding import pad, unpad
from pubnub.exceptions import PubNubException


class PubNubCrypto:
    def __init__(self, pubnub_config):
        self.pubnub_configuration = pubnub_config

    @abstractmethod
    def encrypt(self, key, msg):
        pass

    @abstractmethod
    def decrypt(self, key, msg):
        pass


class CryptoHeader(dict):
    sentinel: str
    header_ver: int = 1
    cryptor_id: str
    cryptor_data: any
    length: any


class CryptorPayload(dict):
    data: bytes
    cryptor_data: bytes


class PubNubCryptor:
    CRYPTOR_ID: str

    @abstractmethod
    def encrypt(self, data: bytes, **kwargs) -> CryptorPayload:
        pass

    @abstractmethod
    def decrypt(self, payload: CryptorPayload, binary_mode: bool = False, **kwargs) -> bytes:
        pass


class PubNubLegacyCryptor(PubNubCryptor):
    CRYPTOR_ID = '0000'
    Initial16bytes = b'0123456789012345'

    def __init__(self, cipher_key, use_random_iv=False, cipher_mode=AES.MODE_CBC, fallback_cipher_mode=None):
        if not cipher_key:
            raise PubNubException('No cipher_key passed')
        self.cipher_key = cipher_key
        self.use_random_iv = use_random_iv
        self.mode = cipher_mode
        self.fallback_mode = fallback_cipher_mode

    def encrypt(self, msg, key=None, use_random_iv=None, **kwargs) -> CryptorPayload:
        key = key or self.cipher_key
        use_random_iv = use_random_iv or self.use_random_iv

        secret = self.get_secret(key)
        initialization_vector = self.get_initialization_vector(use_random_iv)
        cipher = AES.new(bytes(secret[0:32], 'utf-8'), self.mode, initialization_vector)
        encrypted_message = cipher.encrypt(self.pad(msg))
        msg_with_iv = self.append_random_iv(encrypted_message, use_random_iv, initialization_vector)
        return CryptorPayload(data=msg_with_iv, cryptor_data=initialization_vector)

    def decrypt(self, payload: CryptorPayload, key=None, use_random_iv=False, binary_mode: bool = False, **kwargs):
        key = key or self.cipher_key
        use_random_iv = use_random_iv or self.use_random_iv
        secret = self.get_secret(key)
        msg = payload['data']
        initialization_vector, extracted_message = self.extract_random_iv(msg, use_random_iv)
        if not len(extracted_message):
            raise PubNubException('decryption error')
        cipher = AES.new(bytes(secret[0:32], "utf-8"), self.mode, initialization_vector)
        if binary_mode:
            return self.depad(cipher.decrypt(extracted_message), binary_mode)
        try:
            plain = self.depad((cipher.decrypt(extracted_message)).decode('utf-8'), binary_mode)
        except UnicodeDecodeError as e:
            if not self.fallback_mode:
                raise e

            cipher = AES.new(bytes(secret[0:32], "utf-8"), self.fallback_mode, initialization_vector)
            plain = self.depad((cipher.decrypt(extracted_message)).decode('utf-8'), binary_mode)

        try:
            return json.loads(plain)
        except Exception:
            return plain

    def append_random_iv(self, message, use_random_iv, initialization_vector):
        if self.use_random_iv or use_random_iv:
            return initialization_vector + message
        else:
            return message

    def extract_random_iv(self, message, use_random_iv):
        if not isinstance(message, bytes):
            message = bytes(message, 'utf-8')
        if use_random_iv:
            return message[0:16], message[16:]
        else:
            return self.Initial16bytes, message

    def get_initialization_vector(self, use_random_iv) -> bytes:
        if self.use_random_iv or use_random_iv:
            return bytes("{0:016}".format(random.randint(0, 9999999999999999)), 'utf-8')
        else:
            return self.Initial16bytes

    def pad(self, msg, block_size=16):
        padding = block_size - (len(msg) % block_size)
        return msg + (chr(padding) * padding).encode('utf-8')

    def depad(self, msg, binary_mode: bool = False):
        if binary_mode:
            return msg[0:-msg[-1]]
        else:
            return msg[0:-ord(msg[-1])]

    def get_secret(self, key):
        return hashlib.sha256(key.encode("utf-8")).hexdigest()


class PubNubAesCbcCryptor(PubNubCryptor):
    CRYPTOR_ID = 'ACRH'
    mode = AES.MODE_CBC

    def __init__(self, cipher_key):
        self.cipher_key = cipher_key

    def get_initialization_vector(self) -> bytes:
        return secrets.token_bytes(16)

    def get_secret(self, key) -> str:
        return hashlib.sha256(key.encode("utf-8")).digest()

    def encrypt(self, data: bytes, key=None, **kwargs) -> CryptorPayload:
        key = key or self.cipher_key
        secret = self.get_secret(key)
        iv = self.get_initialization_vector()
        cipher = AES.new(secret, mode=self.mode, iv=iv)
        encrypted = cipher.encrypt(pad(data, AES.block_size))
        return CryptorPayload(data=encrypted, cryptor_data=iv)

    def decrypt(self, payload: CryptorPayload, key=None, binary_mode: bool = False, **kwargs):
        key = key or self.cipher_key
        secret = self.get_secret(key)
        iv = payload['cryptor_data']

        cipher = AES.new(secret, mode=self.mode, iv=iv)

        if binary_mode:
            return unpad(cipher.decrypt(payload['data']), AES.block_size)
        else:
            return unpad(cipher.decrypt(payload['data']), AES.block_size).decode()



from pubnub.enums import PNStatusCategory
from pubnub.event_engine.models import invocations
from pubnub.event_engine.models.invocations import PNInvocation
from pubnub.event_engine.models import events
from pubnub.exceptions import PubNubException
from typing import List, Union


class PNContext(dict):
    channels: list
    groups: list
    region: int
    timetoken: str
    attempt: int
    reason: PubNubException
    with_presence: bool = False

    def update(self, context):
        super().update(context.__dict__)


class PNState:
    _context: PNContext

    def __init__(self, context: PNContext) -> None:
        self._context = context
        self._transitions = dict

    def on(self, event: events.PNEvent, context: PNContext):
        return self._transitions[event.get_name()](event, context)

    def on_enter(self, context: Union[None, PNContext]):
        pass

    def on_exit(self):
        pass

    def get_context(self) -> PNContext:
        return self._context


class PNTransition:
    context: PNContext
    state: PNState
    invocation: Union[None, List[PNInvocation]]

    def __init__(self,
                 state: PNState,
                 context: Union[None, PNContext] = None,
                 invocation: Union[None, List[PNInvocation]] = None,
                 ) -> None:
        self.context = context
        self.state = state
        self.invocation = invocation


class UnsubscribedState(PNState):
    def __init__(self, context: PNContext) -> None:
        super().__init__(context)
        self._context.attempt = 0

        self._transitions = {
            events.SubscriptionChangedEvent.__name__: self.subscription_changed,
            events.SubscriptionRestoredEvent.__name__: self.subscription_restored,
        }

    def subscription_changed(self, event: events.SubscriptionChangedEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.channels = event.channels
        self._context.groups = event.groups
        self._context.with_presence = event.with_presence
        self._context.timetoken = 0

        return PNTransition(
            state=HandshakingState,
            context=self._context
        )

    def subscription_restored(self, event: events.SubscriptionRestoredEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.channels = event.channels
        self._context.groups = event.groups
        self._context.with_presence = event.with_presence
        self._context.timetoken = event.timetoken
        self._context.region = event.region

        return PNTransition(
            state=HandshakingState,
            context=self._context
        )


class HandshakingState(PNState):
    def __init__(self, context: PNContext) -> None:
        super().__init__(context)
        self._transitions = {
            events.HandshakeFailureEvent.__name__: self.reconnecting,
            events.DisconnectEvent.__name__: self.disconnect,
            events.HandshakeSuccessEvent.__name__: self.handshaking_success,
            events.SubscriptionRestoredEvent.__name__: self.subscription_restored,
            events.SubscriptionChangedEvent.__name__: self.subscription_changed,
        }

    def on_enter(self, context: Union[None, PNContext]):
        self._context.update(context)
        super().on_enter(self._context)
        return invocations.HandshakeInvocation(self._context.channels,
                                               self._context.groups,
                                               self._context.timetoken or 0)

    def on_exit(self):
        super().on_exit()
        return invocations.CancelHandshakeInvocation()

    def subscription_changed(self, event: events.SubscriptionChangedEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.channels = event.channels
        self._context.groups = event.groups
        self._context.with_presence = event.with_presence
        self._context.timetoken = 0

        return PNTransition(
            state=HandshakingState,
            context=self._context
        )

    def subscription_restored(self, event: events.SubscriptionRestoredEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.channels = event.channels
        self._context.groups = event.groups
        self._context.with_presence = event.with_presence
        self._context.region = event.region
        if self._context.timetoken == 0:
            self._context.timetoken = event.timetoken

        return PNTransition(
            state=ReceivingState,
            context=self._context
        )

    def reconnecting(self, event: events.HandshakeFailureEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.attempt = event.attempt
        self._context.reason = event.reason
        if self._context.timetoken == 0:
            self._context.timetoken = event.timetoken

        return PNTransition(
            state=HandshakeReconnectingState,
            context=self._context
        )

    def disconnect(self, event: events.DisconnectEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.timetoken = 0

        return PNTransition(
            state=HandshakeStoppedState,
            context=self._context
        )

    def handshaking_success(self, event: events.HandshakeSuccessEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.timetoken = event.timetoken
        self._context.region = event.region
        self._context.attempt = 0

        return PNTransition(
            state=ReceivingState,
            context=self._context,
            invocation=invocations.EmitStatusInvocation(PNStatusCategory.PNConnectedCategory)
        )


class HandshakeReconnectingState(PNState):
    def __init__(self, context: PNContext) -> None:
        super().__init__(context)
        self._transitions = {
            events.DisconnectEvent.__name__: self.disconnect,
            events.HandshakeReconnectGiveupEvent.__name__: self.give_up,
            events.HandshakeReconnectSuccessEvent.__name__: self.success,
            events.SubscriptionRestoredEvent.__name__: self.subscription_restored,
            events.HandshakeReconnectFailureEvent.__name__: self.handshake_reconnect,
            events.SubscriptionChangedEvent.__name__: self.subscription_changed,
        }

    def on_enter(self, context: Union[None, PNContext]):
        self._context.update(context)
        super().on_enter(self._context)
        return invocations.HandshakeReconnectInvocation(self._context.channels,
                                                        self._context.groups,
                                                        attempts=self._context.attempt,
                                                        reason=self._context.reason,
                                                        timetoken=self._context.timetoken)

    def on_exit(self):
        super().on_exit()
        return invocations.CancelHandshakeReconnectInvocation()

    def disconnect(self, event: events.DisconnectEvent, context: PNContext) -> PNTransition:
        self._context.update(context)

        return PNTransition(
            state=HandshakeStoppedState,
            context=self._context
        )

    def subscription_changed(self, event: events.SubscriptionChangedEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.channels = event.channels
        self._context.groups = event.groups
        self._context.with_presence = event.with_presence
        self._context.timetoken = 0

        return PNTransition(
            state=HandshakeReconnectingState,
            context=self._context
        )

    def handshake_reconnect(self, event: events.HandshakeReconnectFailureEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.attempt = event.attempt + 1
        self._context.reason = event.reason

        return PNTransition(
            state=HandshakeReconnectingState,
            context=self._context
        )

    def give_up(self, event: events.HandshakeReconnectGiveupEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.attempt = event.attempt
        self._context.reason = event.reason

        return PNTransition(
            state=HandshakeFailedState,
            context=self._context,
            invocation=invocations.EmitStatusInvocation(PNStatusCategory.PNDisconnectedCategory)
        )

    def subscription_restored(self, event: events.SubscriptionRestoredEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.channels = event.channels
        self._context.groups = event.groups
        self._context.with_presence = event.with_presence
        self._context.timetoken = event.timetoken
        self._context.region = event.region

        return PNTransition(
            state=ReceivingState,
            context=self._context
        )

    def success(self, event: events.HandshakeReconnectSuccessEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.timetoken = event.timetoken
        self._context.region = event.region

        return PNTransition(
            state=ReceivingState,
            context=self._context,
            invocation=invocations.EmitStatusInvocation(PNStatusCategory.PNConnectedCategory, )
        )


class HandshakeFailedState(PNState):
    def __init__(self, context: PNContext) -> None:
        super().__init__(context)
        self._transitions = {
            events.SubscriptionChangedEvent.__name__: self.subscription_changed,
            events.ReconnectEvent.__name__: self.reconnect,
            events.SubscriptionRestoredEvent.__name__: self.subscription_restored,
        }

    def subscription_changed(self, event: events.SubscriptionChangedEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.channels = event.channels
        self._context.groups = event.groups
        self._context.with_presence = event.with_presence
        self._context.timetoken = 0

        return PNTransition(
            state=HandshakingState,
            context=self._context
        )

    def reconnect(self, event: events.ReconnectEvent, context: PNContext) -> PNTransition:
        self._context.update(context)

        return PNTransition(
            state=HandshakingState,
            context=self._context
        )

    def subscription_restored(self, event: events.SubscriptionRestoredEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.channels = event.channels
        self._context.groups = event.groups
        self._context.with_presence = event.with_presence
        self._context.timetoken = event.timetoken
        self._context.region = event.region

        return PNTransition(
            state=ReceivingState,
            context=self._context
        )


class HandshakeStoppedState(PNState):
    def __init__(self, context: PNContext) -> None:
        super().__init__(context)
        self._context.attempt = 0

        self._transitions = {
            events.ReconnectEvent.__name__: self.reconnect
        }

    def reconnect(self, event: events.ReconnectEvent, context: PNContext) -> PNTransition:
        self._context.update(context)

        return PNTransition(
            state=HandshakeReconnectingState,
            context=self._context
        )


class ReceivingState(PNState):
    def __init__(self, context: PNContext) -> None:
        super().__init__(context)
        self._context.attempt = 0

        self._transitions = {
            events.SubscriptionChangedEvent.__name__: self.subscription_changed,
            events.SubscriptionRestoredEvent.__name__: self.subscription_restored,
            events.ReceiveSuccessEvent.__name__: self.receiving_success,
            events.ReceiveFailureEvent.__name__: self.receiving_failure,
            events.DisconnectEvent.__name__: self.disconnect,
            events.ReconnectEvent.__name__: self.reconnect,
        }

    def on_enter(self, context: Union[None, PNContext]):
        super().on_enter(context)
        return invocations.ReceiveMessagesInvocation(context.channels, context.groups,
                                                     timetoken=self._context.timetoken, region=self._context.region)

    def on_exit(self):
        super().on_exit()
        return invocations.CancelReceiveMessagesInvocation()

    def subscription_changed(self, event: events.SubscriptionChangedEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.channels = event.channels
        self._context.groups = event.groups
        self._context.with_presence = event.with_presence
        # self._context.timetoken = 0  # why we don't reset timetoken here?

        return PNTransition(
            state=self.__class__,
            context=self._context
        )

    def subscription_restored(self, event: events.SubscriptionRestoredEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.channels = event.channels
        self._context.groups = event.groups
        self._context.with_presence = event.with_presence
        self._context.timetoken = event.timetoken
        self._context.region = event.region

        return PNTransition(
            state=self.__class__,
            context=self._context
        )

    def receiving_success(self, event: events.ReceiveSuccessEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.timetoken = event.timetoken
        self._context.region = event.region

        return PNTransition(
            state=self.__class__,
            context=self._context,
            invocation=invocations.EmitMessagesInvocation(messages=event.messages),
        )

    def receiving_failure(self, event: events.ReceiveFailureEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.reason = event.reason
        self._context.attempt = event.attempt
        self._context.timetoken = event.timetoken
        return PNTransition(
            state=ReceiveReconnectingState,
            context=self._context
        )

    def disconnect(self, event: events.DisconnectEvent, context: PNContext) -> PNTransition:
        self._context.update(context)

        return PNTransition(
            state=ReceiveStoppedState,
            context=self._context,
            invocation=invocations.EmitStatusInvocation(PNStatusCategory.PNDisconnectedCategory)
        )

    def reconnect(self, event: events.ReconnectEvent, context: PNContext) -> PNTransition:
        self._context.update(context)

        return PNTransition(
            state=ReceivingState,
            context=self._context
        )


class ReceiveReconnectingState(PNState):
    def __init__(self, context: PNContext) -> None:
        super().__init__(context)
        self._transitions = {
            events.ReceiveReconnectFailureEvent.__name__: self.reconnect_failure,
            events.SubscriptionChangedEvent.__name__: self.subscription_changed,
            events.DisconnectEvent.__name__: self.disconnect,
            events.ReceiveReconnectGiveupEvent.__name__: self.give_up,
            events.ReceiveReconnectSuccessEvent.__name__: self.reconnect_success,
            events.SubscriptionRestoredEvent.__name__: self.subscription_restored,
        }

    def on_enter(self, context: Union[None, PNContext]):
        self._context.update(context)
        super().on_enter(self._context)
        return invocations.ReceiveReconnectInvocation(self._context.channels,
                                                      self._context.groups,
                                                      self._context.timetoken,
                                                      self._context.region,
                                                      self._context.attempt,
                                                      self._context.reason)

    def on_exit(self):
        super().on_exit()
        return invocations.CancelReceiveReconnectInvocation()

    def reconnect_failure(self, event: events.ReceiveReconnectFailureEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.attempt = event.attempt + 1
        self._context.reason = event.reason

        return PNTransition(
            state=ReceiveReconnectingState,
            context=self._context
        )

    def subscription_changed(self, event: events.SubscriptionChangedEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.channels = event.channels
        self._context.groups = event.groups
        self._context.with_presence = event.with_presence
        self._context.timetoken = 0

        return PNTransition(
            state=ReceiveReconnectingState,
            context=self._context
        )

    def disconnect(self, event: events.DisconnectEvent, context: PNContext) -> PNTransition:
        self._context.update(context)

        return PNTransition(
            state=ReceiveStoppedState,
            context=self._context
        )

    def give_up(self, event: events.ReceiveReconnectGiveupEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.reason = event.reason
        self._context.attempt = event.attempt

        return PNTransition(
            state=ReceiveFailedState,
            context=self._context,
            invocation=invocations.EmitStatusInvocation(PNStatusCategory.PNDisconnectedCategory)
        )

    def reconnect_success(self, event: events.ReceiveReconnectSuccessEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.timetoken = event.timetoken
        self._context.region = event.region

        return PNTransition(
            state=ReceivingState,
            context=self._context,
            invocation=invocations.EmitMessagesInvocation(event.messages)
        )

    def subscription_restored(self, event: events.SubscriptionRestoredEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.channels = event.channels
        self._context.groups = event.groups
        self._context.with_presence = event.with_presence
        self._context.timetoken = event.timetoken
        self._context.region = event.region

        return PNTransition(
            state=ReceiveReconnectingState,
            context=self._context
        )


class ReceiveFailedState(PNState):
    def __init__(self, context: PNContext) -> None:
        super().__init__(context)
        self._transitions = {
            events.ReceiveReconnectRetryEvent.__name__: self.reconnect_retry,
            events.SubscriptionChangedEvent.__name__: self.subscription_changed,
            events.SubscriptionRestoredEvent.__name__: self.subscription_restored,
            events.ReconnectEvent.__name__: self.reconnect,
        }

    def reconnect_retry(self, event: events.ReceiveReconnectRetryEvent, context: PNContext) -> PNTransition:
        self._context.update(context)

        return PNTransition(
            state=ReceiveReconnectingState,
            context=self._context
        )

    def subscription_changed(self, event: events.SubscriptionChangedEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.channels = event.channels
        self._context.groups = event.groups
        self._context.with_presence = event.with_presence
        self._context.timetoken = 0

        return PNTransition(
            state=ReceivingState,
            context=self._context
        )

    def reconnect(self, event: events.ReconnectEvent, context: PNContext) -> PNTransition:
        self._context.update(context)

        return PNTransition(
            state=ReceivingState,
            context=self._context
        )

    def subscription_restored(self, event: events.SubscriptionRestoredEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.channels = event.channels
        self._context.groups = event.groups
        self._context.with_presence = event.with_presence
        self._context.timetoken = event.timetoken
        self._context.region = event.region

        return PNTransition(
            state=ReceivingState,
            context=self._context
        )


class ReceiveStoppedState(PNState):
    def __init__(self, context: PNContext) -> None:
        super().__init__(context)
        self._context.attempt = 0

        self._transitions = {
            events.ReconnectEvent.__name__: self.reconnect
        }

    def reconnect(self, event: events.ReconnectEvent, context: PNContext) -> PNTransition:
        self._context.update(context)

        return PNTransition(
            state=ReceiveReconnectingState,
            context=self._context
        )


"""
Presence states
"""


class HeartbeatInactiveState(PNState):
    def __init__(self, context: PNContext) -> None:
        super().__init__(context)

        self._transitions = {
            events.HeartbeatJoinedEvent.__name__: self.joined
        }

    def joined(self, event: events.HeartbeatJoinedEvent, context: PNContext) -> PNTransition:
        self._context.channels = event.channels
        self._context.groups = event.groups
        self._context.update(context)

        return PNTransition(
            state=HeartbeatingState,
            context=self._context
        )


class HeartbeatStoppedState(PNState):
    def __init__(self, context: PNContext) -> None:
        super().__init__(context)

        self._transitions = {
            events.HeartbeatReconnectEvent.__name__: self.reconnect,
            events.HeartbeatLeftAllEvent.__name__: self.left_all,
            events.HeartbeatJoinedEvent.__name__: self.joined,
            events.HeartbeatLeftEvent.__name__: self.left
        }

    def reconnect(self, event: events.HeartbeatReconnectEvent, context: PNContext) -> PNTransition:
        self._context.update(context)

        return PNTransition(
            state=HeartbeatingState,
            context=self._context
        )

    def left_all(self, event: events.HeartbeatLeftAllEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.channels = []
        self._context.groups = []

        return PNTransition(
            state=HeartbeatInactiveState,
            context=self._context
        )

    def joined(self, event: events.HeartbeatJoinedEvent, context: PNContext) -> PNTransition:
        self._context.update(context)

        return PNTransition(
            state=HeartbeatStoppedState,
            context=self._context
        )

    def left(self, event: events.HeartbeatLeftEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        for channel in event.channels:
            self._context.channels.remove(channel)

        for group in event.groups:
            self._context.groups.remove(group)

        return PNTransition(
            state=HeartbeatStoppedState,
            context=self._context
        )


class HeartbeatFailedState(PNState):
    def __init__(self, context: PNContext) -> None:
        super().__init__(context)

        self._transitions = {
            events.HeartbeatJoinedEvent.__name__: self.joined,
            events.HeartbeatLeftEvent.__name__: self.left,
            events.HeartbeatReconnectEvent.__name__: self.reconnect,
            events.HeartbeatDisconnectEvent.__name__: self.disconnect,
            events.HeartbeatLeftAllEvent.__name__: self.left_all
        }

    def joined(self, event: events.HeartbeatJoinedEvent, context: PNContext) -> PNTransition:
        self._context.update(context)

        return PNTransition(
            state=HeartbeatingState,
            context=self._context
        )

    def left(self, event: events.HeartbeatLeftEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        for channel in event.channels:
            self._context.channels.remove(channel)

        for group in event.groups:
            self._context.groups.remove(group)

        invocation = None
        if not event.suppress_leave:
            invocation = invocations.HeartbeatLeaveInvocation(channels=event.channels,
                                                              groups=event.groups)

        return PNTransition(
            state=HeartbeatingState,
            context=self._context,
            invocation=invocation
        )

    def reconnect(self, event: events.HeartbeatReconnectEvent, context: PNContext) -> PNTransition:
        self._context.update(context)

        return PNTransition(
            state=HeartbeatingState,
            context=self._context
        )

    def disconnect(self, event: events.HeartbeatDisconnectEvent, context: PNContext) -> PNTransition:
        self._context.update(context)

        invocation = None
        if not event.suppress_leave:
            invocation = invocations.HeartbeatLeaveInvocation(channels=event.channels,
                                                              groups=event.groups)

        return PNTransition(
            state=HeartbeatStoppedState,
            context=self._context,
            invocation=invocation
        )

    def left_all(self, event: events.HeartbeatLeftAllEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.channels = []
        self._context.groups = []

        invocation = None
        if not event.suppress_leave:
            invocation = invocations.HeartbeatLeaveInvocation(channels=event.channels,
                                                              groups=event.groups)

        return PNTransition(
            state=HeartbeatInactiveState,
            context=self._context,
            invocation=invocation
        )


class HeartbeatingState(PNState):
    def __init__(self, context: PNContext) -> None:
        super().__init__(context)
        self._transitions = {
            events.HeartbeatFailureEvent.__name__: self.failure,
            events.HeartbeatDisconnectEvent.__name__: self.disconnect,
            events.HeartbeatLeftAllEvent.__name__: self.left_all,
            events.HeartbeatJoinedEvent.__name__: self.joined,
            events.HeartbeatLeftEvent.__name__: self.left,
            events.HeartbeatSuccessEvent.__name__: self.success
        }

    def on_enter(self, context: Union[None, PNContext]):
        self._context.update(context)
        super().on_enter(self._context)
        return invocations.HeartbeatInvocation(channels=self._context.channels, groups=self._context.groups)

    def failure(self, event: events.HeartbeatFailureEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.attempt = event.attempt
        self._context.reason = event.reason

        return PNTransition(
            state=HeartbeatReconnectingState,
            context=self._context
        )

    def disconnect(self, event: events.HeartbeatDisconnectEvent, context: PNContext) -> PNTransition:
        self._context.update(context)

        invocation = None
        if not event.suppress_leave:
            invocation = invocations.HeartbeatLeaveInvocation(channels=event.channels,
                                                              groups=event.groups)

        return PNTransition(
            state=HeartbeatStoppedState,
            context=self._context,
            invocation=invocation
        )

    def left_all(self, event: events.HeartbeatLeftAllEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.channels = []
        self._context.groups = []

        invocation = None
        if not event.suppress_leave:
            invocation = invocations.HeartbeatLeaveInvocation(channels=event.channels,
                                                              groups=event.groups)

        return PNTransition(
            state=HeartbeatInactiveState,
            context=self._context,
            invocation=invocation
        )

    def joined(self, event: events.HeartbeatJoinedEvent, context: PNContext) -> PNTransition:
        self._context.update(context)

        return PNTransition(
            state=HeartbeatingState,
            context=self._context
        )

    def left(self, event: events.HeartbeatLeftEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        for channel in event.channels:
            self._context.channels.remove(channel)

        for group in event.groups:
            self._context.groups.remove(group)

        invocation = None
        if not event.suppress_leave:
            invocation = invocations.HeartbeatLeaveInvocation(channels=event.channels,
                                                              groups=event.groups)

        return PNTransition(
            state=HeartbeatingState,
            context=self._context,
            invocation=invocation
        )

    def success(self, event: events.HeartbeatSuccessEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.attempt = 0

        return PNTransition(
            state=HeartbeatCooldownState,
            context=self._context
        )


class HeartbeatCooldownState(PNState):
    def __init__(self, context: PNContext) -> None:
        super().__init__(context)
        self._transitions = {
            events.HeartbeatJoinedEvent.__name__: self.joined,
            events.HeartbeatLeftEvent.__name__: self.left,
            events.HeartbeatTimesUpEvent.__name__: self.times_up,
            events.HeartbeatDisconnectEvent.__name__: self.disconnect,
            events.HeartbeatLeftAllEvent.__name__: self.left_all,

        }

    def on_enter(self, context: PNContext):
        self._context.update(context)
        super().on_enter(self._context)
        return invocations.HeartbeatWaitInvocation(self._context)

    def on_exit(self):
        super().on_exit()
        return invocations.HeartbeatCancelWaitInvocation()

    def disconnect(self, event: events.HeartbeatDisconnectEvent, context: PNContext) -> PNTransition:
        self._context.update(context)

        invocation = None
        if not event.suppress_leave:
            invocation = invocations.HeartbeatLeaveInvocation(channels=event.channels,
                                                              groups=event.groups)

        return PNTransition(
            state=HeartbeatStoppedState,
            context=self._context,
            invocation=invocation
        )

    def left_all(self, event: events.HeartbeatLeftAllEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.channels = []
        self._context.groups = []

        invocation = None
        if not event.suppress_leave:
            invocation = invocations.HeartbeatLeaveInvocation(channels=event.channels,
                                                              groups=event.groups)

        return PNTransition(
            state=HeartbeatInactiveState,
            context=self._context,
            invocation=invocation
        )

    def joined(self, event: events.HeartbeatJoinedEvent, context: PNContext) -> PNTransition:
        self._context.update(context)

        return PNTransition(
            state=HeartbeatingState,
            context=self._context
        )

    def left(self, event: events.HeartbeatLeftEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        for channel in event.channels:
            self._context.channels.remove(channel)

        for group in event.groups:
            self._context.groups.remove(group)

        invocation = None
        if not event.suppress_leave:
            invocation = invocations.HeartbeatLeaveInvocation(channels=event.channels,
                                                              groups=event.groups)

        return PNTransition(
            state=HeartbeatingState,
            context=self._context,
            invocation=invocation
        )

    def times_up(self, event: events.HeartbeatTimesUpEvent, context: PNContext) -> PNTransition:
        self._context.update(context)

        return PNTransition(
            state=HeartbeatingState,
            context=self._context
        )


class HeartbeatReconnectingState(PNState):
    def __init__(self, context: PNContext) -> None:
        super().__init__(context)
        self._transitions = {
            events.HeartbeatFailureEvent.__name__: self.failure,
            events.HeartbeatJoinedEvent.__name__: self.joined,
            events.HeartbeatLeftEvent.__name__: self.left,
            events.HeartbeatSuccessEvent.__name__: self.success,
            events.HeartbeatGiveUpEvent.__name__: self.give_up,
            events.HeartbeatDisconnectEvent.__name__: self.disconnect,
            events.HeartbeatLeftAllEvent.__name__: self.left_all
        }

    def on_enter(self, context: PNContext):
        self._context.update(context)
        super().on_enter(self._context)

        return invocations.HeartbeatDelayedHeartbeatInvocation(channels=self._context.channels,
                                                               groups=self._context.groups,
                                                               attempts=self._context.attempt,
                                                               reason=None)

    def on_exit(self):
        super().on_exit()
        return invocations.HeartbeatCancelDelayedHeartbeatInvocation()

    def failure(self, event: events.HeartbeatFailureEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.attempt = event.attempt + 1
        self._context.reason = event.reason

        return PNTransition(
            state=HeartbeatReconnectingState,
            context=self._context
        )

    def joined(self, event: events.HeartbeatJoinedEvent, context: PNContext) -> PNTransition:
        self._context.update(context)

        return PNTransition(
            state=HeartbeatingState,
            context=self._context
        )

    def left(self, event: events.HeartbeatLeftEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        for channel in event.channels:
            self._context.channels.remove(channel)

        for group in event.groups:
            self._context.groups.remove(group)

        invocation = None
        if not event.suppress_leave:
            invocation = invocations.HeartbeatLeaveInvocation(channels=event.channels,
                                                              groups=event.groups)

        return PNTransition(
            state=HeartbeatingState,
            context=self._context,
            invocation=invocation
        )

    def success(self, event: events.HeartbeatSuccessEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.attempt = 0

        return PNTransition(
            state=HeartbeatCooldownState,
            context=self._context
        )

    def give_up(self, event: events.HeartbeatGiveUpEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.attempt = event.attempt
        self._context.reason = event.reason

        return PNTransition(
            state=HeartbeatFailedState,
            context=self._context
        )

    def disconnect(self, event: events.HeartbeatDisconnectEvent, context: PNContext) -> PNTransition:
        self._context.update(context)

        invocation = None
        if not event.suppress_leave:
            invocation = invocations.HeartbeatLeaveInvocation(channels=event.channels,
                                                              groups=event.groups)

        return PNTransition(
            state=HeartbeatStoppedState,
            context=self._context,
            invocation=invocation
        )

    def left_all(self, event: events.HeartbeatLeftAllEvent, context: PNContext) -> PNTransition:
        self._context.update(context)
        self._context.channels = []
        self._context.groups = []

        invocation = None
        if not event.suppress_leave:
            invocation = invocations.HeartbeatLeaveInvocation(channels=event.channels,
                                                              groups=event.groups)

        return PNTransition(
            state=HeartbeatInactiveState,
            context=self._context,
            invocation=invocation
        )



from pubnub.endpoints.file_operations.file_based_endpoint import FileOperationEndpoint
from pubnub.enums import HttpMethod, PNOperationType
from pubnub import utils
from pubnub.models.consumer.file import PNPublishFileMessageResult
from pubnub.endpoints.mixins import TimeTokenOverrideMixin
from pubnub.crypto import PubNubCryptodome
from warnings import warn


class PublishFileMessage(FileOperationEndpoint, TimeTokenOverrideMixin):
    PUBLISH_FILE_MESSAGE = "/v1/files/publish-file/%s/%s/0/%s/0/%s"

    def __init__(self, pubnub):
        super(PublishFileMessage, self).__init__(pubnub)
        self._file_id = None
        self._file_name = None
        self._pubnub = pubnub
        self._message = None
        self._should_store = None
        self._ttl = 0
        self._meta = None
        self._cipher_key = None
        self._replicate = None
        self._ptto = None

    def meta(self, meta):
        self._meta = meta
        return self

    def should_store(self, should_store):
        self._should_store = bool(should_store)
        return self

    def cipher_key(self, cipher_key):
        if cipher_key:
            warn('Deprecated: Usage of local cipher_keys is discouraged. Use pnconfiguration.cipher_key instead')
            self._cipher_key = cipher_key
        return self

    def message(self, message):
        self._message = message
        return self

    def file_id(self, file_id):
        self._file_id = file_id
        return self

    def ttl(self, ttl):
        self._ttl = ttl
        return self

    def file_name(self, file_name):
        self._file_name = file_name
        return self

    def _encrypt_message(self, message):
        if self._cipher_key:
            return PubNubCryptodome(self._pubnub.config).encrypt(self._cipher_key, utils.write_value_as_string(message))
        elif self._pubnub.config.cipher_key:
            return self._pubnub.crypto.encrypt(utils.write_value_as_string(message))
        else:
            return message

    def _build_message(self):
        message = {
            "message": self._message,
            "file": {
                "id": self._file_id,
                "name": self._file_name
            }
        }
        return self._encrypt_message(message)

    def build_path(self):
        message = self._build_message()
        return PublishFileMessage.PUBLISH_FILE_MESSAGE % (
            self.pubnub.config.publish_key,
            self.pubnub.config.subscribe_key,
            utils.url_encode(self._channel),
            utils.url_write(message)
        )

    def http_method(self):
        return HttpMethod.GET

    def custom_params(self):
        params = TimeTokenOverrideMixin.custom_params(self)
        params.update({
            "meta": utils.url_write(self._meta),
            "ttl": self._ttl,
            "store": 1 if self._should_store else 0
        })
        return params

    def is_auth_required(self):
        return True

    def validate_params(self):
        self.validate_subscribe_key()
        self.validate_channel()
        self.validate_file_name()
        self.validate_file_id()

    def create_response(self, envelope):
        return PNPublishFileMessageResult(envelope)

    def operation_type(self):
        return PNOperationType.PNSendFileAction

    def name(self):
        return "Sending file upload notification"



from pubnub.exceptions import PubNubException
from typing import List, Optional


class PNEvent:
    def get_name(self) -> str:
        return self.__class__.__name__


class PNFailureEvent(PNEvent):
    def __init__(self, reason: PubNubException, attempt: int, timetoken: int = 0) -> None:
        self.reason = reason
        self.attempt = attempt
        self.timetoken = timetoken
        super().__init__()


class PNCursorEvent(PNEvent):
    def __init__(self, timetoken: str, region: Optional[int] = None, **kwargs) -> None:
        self.timetoken = timetoken
        self.region = region


class PNChannelGroupsEvent(PNEvent):
    def __init__(self, channels: List[str], groups: List[str]) -> None:
        self.channels = channels
        self.groups = groups


class SubscriptionChangedEvent(PNChannelGroupsEvent):
    def __init__(self, channels: List[str], groups: List[str], with_presence: Optional[bool] = None) -> None:
        PNChannelGroupsEvent.__init__(self, channels, groups)
        self.with_presence = with_presence


class SubscriptionRestoredEvent(PNCursorEvent, PNChannelGroupsEvent):
    def __init__(self, timetoken: str, channels: List[str], groups: List[str], region: Optional[int] = None,
                 with_presence: Optional[bool] = None) -> None:
        PNCursorEvent.__init__(self, timetoken, region)
        PNChannelGroupsEvent.__init__(self, channels, groups)
        self.with_presence = with_presence


class HandshakeSuccessEvent(PNCursorEvent):
    def __init__(self, timetoken: str, region: Optional[int] = None, **kwargs) -> None:
        super().__init__(timetoken, region)


class HandshakeFailureEvent(PNFailureEvent):
    pass


class HandshakeReconnectSuccessEvent(PNCursorEvent):
    pass


class HandshakeReconnectFailureEvent(PNFailureEvent):
    pass


class HandshakeReconnectGiveupEvent(PNFailureEvent):
    pass


class HandshakeReconnectRetryEvent(PNEvent):
    pass


class ReceiveSuccessEvent(PNCursorEvent):
    def __init__(self, timetoken: str, messages: list, region: Optional[int] = None) -> None:
        PNCursorEvent.__init__(self, timetoken, region)
        self.messages = messages


class ReceiveFailureEvent(PNFailureEvent):
    pass


class ReceiveReconnectSuccessEvent(PNCursorEvent):
    def __init__(self, timetoken: str, messages: list, region: Optional[int] = None) -> None:
        PNCursorEvent.__init__(self, timetoken, region)
        self.messages = messages


class ReceiveReconnectFailureEvent(PNFailureEvent):
    pass


class ReceiveReconnectGiveupEvent(PNFailureEvent):
    pass


class ReceiveReconnectRetryEvent(PNEvent):
    pass


class DisconnectEvent(PNEvent):
    pass


class ReconnectEvent(PNEvent):
    pass


"""
    Presence Events
"""


class HeartbeatJoinedEvent(PNChannelGroupsEvent):
    pass


class HeartbeatReconnectEvent(PNEvent):
    pass


class HeartbeatLeftAllEvent(PNEvent):
    pass


class HeartbeatLeftEvent(PNChannelGroupsEvent):
    def __init__(self, channels: List[str], groups: List[str], suppress_leave: bool = False) -> None:
        PNChannelGroupsEvent.__init__(self, channels, groups)
        self.suppress_leave = suppress_leave


class HeartbeatDisconnectEvent(PNChannelGroupsEvent):
    pass


class HeartbeatSuccessEvent(PNChannelGroupsEvent):
    pass


class HeartbeatFailureEvent(PNChannelGroupsEvent, PNFailureEvent):
    def __init__(self, channels: List[str], groups: List[str], reason: PubNubException, attempt: int,
                 timetoken: int = 0) -> None:
        PNChannelGroupsEvent.__init__(self, channels, groups)
        PNFailureEvent.__init__(self, reason, attempt, timetoken)


class HeartbeatTimesUpEvent(PNEvent):
    pass


class HeartbeatGiveUpEvent(PNChannelGroupsEvent, PNFailureEvent):
    def __init__(self, channels: List[str], groups: List[str], reason: PubNubException, attempt: int,
                 timetoken: int = 0) -> None:
        PNChannelGroupsEvent.__init__(self, channels, groups)
        PNFailureEvent.__init__(self, reason, attempt, timetoken)



import asyncio
import logging
import math

from typing import Optional, Union
from pubnub.endpoints.presence.heartbeat import Heartbeat
from pubnub.endpoints.presence.leave import Leave
from pubnub.endpoints.pubsub.subscribe import Subscribe
from pubnub.enums import PNReconnectionPolicy
from pubnub.exceptions import PubNubException
from pubnub.features import feature_enabled
from pubnub.models.server.subscribe import SubscribeMessage
from pubnub.pubnub import PubNub
from pubnub.event_engine.models import events, invocations
from pubnub.models.consumer.common import PNStatus
from pubnub.workers import BaseMessageWorker


class Effect:
    pubnub: PubNub = None
    event_engine = None
    invocation: Union[invocations.PNManageableInvocation, invocations.PNCancelInvocation]
    stop_event = None
    logger: logging.Logger
    task: asyncio.Task

    def set_pn(self, pubnub: PubNub):
        self.pubnub = pubnub

    def __init__(self, pubnub_instance, event_engine_instance,
                 invocation: Union[invocations.PNManageableInvocation, invocations.PNCancelInvocation]) -> None:
        self.invocation = invocation
        self.event_engine = event_engine_instance
        self.pubnub = pubnub_instance

        self.logger = logging.getLogger("pubnub")

    def run(self):
        pass

    def run_async(self, coro):
        loop: asyncio.AbstractEventLoop = self.pubnub.event_loop
        if loop.is_running():
            self.task = loop.create_task(coro, name=self.__class__.__name__)
        else:
            self.task = loop.run_until_complete(coro)

    def stop(self):
        if self.stop_event:
            self.logger.debug(f'stop_event({id(self.stop_event)}).set() called on {self.__class__.__name__}')
            self.stop_event.set()
        if hasattr(self, 'task') and isinstance(self.task, asyncio.Task) and not self.task.cancelled():
            self.task.cancel()

    def get_new_stop_event(self):
        event = asyncio.Event()
        self.logger.debug(f'creating new stop_event({id(event)}) for {self.__class__.__name__}')
        return event

    def calculate_reconnection_delay(self, attempts):
        if self.reconnection_policy is PNReconnectionPolicy.LINEAR:
            delay = self.interval

        elif self.reconnection_policy is PNReconnectionPolicy.EXPONENTIAL:
            delay = int(math.pow(2, attempts - 5 * math.floor((attempts - 1) / 5)) - 1)
        return delay


class HandshakeEffect(Effect):
    def run(self):
        channels = self.invocation.channels
        groups = self.invocation.groups
        tt = self.invocation.timetoken or 0
        if hasattr(self.pubnub, 'event_loop'):
            self.stop_event = self.get_new_stop_event()
            self.run_async(self.handshake_async(channels=channels,
                                                groups=groups,
                                                timetoken=tt,
                                                stop_event=self.stop_event))

    async def handshake_async(self, channels, groups, stop_event, timetoken: int = 0):
        request = Subscribe(self.pubnub).channels(channels).channel_groups(groups).cancellation_event(stop_event)

        if feature_enabled('PN_MAINTAIN_PRESENCE_STATE') and hasattr(self.pubnub, 'state_container'):
            state_container = self.pubnub.state_container
            request.state(state_container.get_state(channels))

        request.timetoken(0)
        response = await request.future()

        if isinstance(response, PubNubException):
            self.logger.warning(f'Handshake failed: {str(response)}')
            handshake_failure = events.HandshakeFailureEvent(str(response), 1, timetoken=timetoken)
            self.event_engine.trigger(handshake_failure)
        elif response.status.error:
            self.logger.warning(f'Handshake failed: {response.status.error_data.__dict__}')
            handshake_failure = events.HandshakeFailureEvent(response.status.error_data, 1, timetoken=timetoken)
            self.event_engine.trigger(handshake_failure)
        else:
            cursor = response.result['t']
            timetoken = timetoken if timetoken > 0 else cursor['t']
            region = cursor['r']
            handshake_success = events.HandshakeSuccessEvent(timetoken, region)
            self.event_engine.trigger(handshake_success)


class ReceiveMessagesEffect(Effect):
    invocation: invocations.ReceiveMessagesInvocation

    def run(self):
        channels = self.invocation.channels
        groups = self.invocation.groups
        timetoken = self.invocation.timetoken
        region = self.invocation.region

        if hasattr(self.pubnub, 'event_loop'):
            self.stop_event = self.get_new_stop_event()
            self.run_async(self.receive_messages_async(channels, groups, timetoken, region))

    async def receive_messages_async(self, channels, groups, timetoken, region):
        request = Subscribe(self.pubnub)
        if channels:
            request.channels(channels)
        if groups:
            request.channel_groups(groups)
        if timetoken:
            request.timetoken(timetoken)
        if region:
            request.region(region)

        request.cancellation_event(self.stop_event)
        response = await request.future()

        if response.status is None and response.result is None:
            self.logger.warning('Recieve messages failed: Empty response')
            recieve_failure = events.ReceiveFailureEvent('Empty response', 1, timetoken=timetoken)
            self.event_engine.trigger(recieve_failure)
        elif response.status.error:
            self.logger.warning(f'Recieve messages failed: {response.status.error_data.__dict__}')
            recieve_failure = events.ReceiveFailureEvent(response.status.error_data, 1, timetoken=timetoken)
            self.event_engine.trigger(recieve_failure)
        else:
            cursor = response.result['t']
            timetoken = cursor['t']
            region = cursor['r']
            messages = response.result['m']
            recieve_success = events.ReceiveSuccessEvent(timetoken, region=region, messages=messages)
            self.event_engine.trigger(recieve_success)
        self.stop_event.set()


class ReconnectEffect(Effect):
    invocation: invocations.ReconnectInvocation
    reconnection_policy: PNReconnectionPolicy

    def __init__(self, pubnub_instance, event_engine_instance,
                 invocation: Union[invocations.PNManageableInvocation, invocations.PNCancelInvocation]) -> None:
        super().__init__(pubnub_instance, event_engine_instance, invocation)
        self.reconnection_policy = pubnub_instance.config.reconnect_policy
        self.max_retry_attempts = pubnub_instance.config.maximum_reconnection_retries
        self.interval = pubnub_instance.config.RECONNECTION_INTERVAL
        self.min_backoff = pubnub_instance.config.RECONNECTION_MIN_EXPONENTIAL_BACKOFF
        self.max_backoff = pubnub_instance.config.RECONNECTION_MAX_EXPONENTIAL_BACKOFF

    def give_up(self, reason: PubNubException, attempt: int, timetoken: int = 0):
        self.logger.error(f"GiveUp called on Unspecific event. Reason: {reason}, Attempt: {attempt} TT:{timetoken}")
        raise PubNubException('Unspecified Invocation')

    def failure(self, reason: PubNubException, attempt: int, timetoken: int = 0):
        self.logger.error(f"Failure called on Unspecific event. Reason: {reason}, Attempt: {attempt} TT:{timetoken}")
        raise PubNubException('Unspecified Invocation')

    def success(self, timetoken: str, region: Optional[int] = None, **kwargs):
        self.logger.error(f"Success called on Unspecific event. TT:{timetoken}, Reg: {region}, KWARGS: {kwargs.keys()}")
        raise PubNubException('Unspecified Invocation')

    def run(self):
        if self.reconnection_policy is PNReconnectionPolicy.NONE or self.invocation.attempts > self.max_retry_attempts:
            self.give_up(reason=self.invocation.reason, attempt=self.invocation.attempts)
        else:
            attempts = self.invocation.attempts
            delay = self.calculate_reconnection_delay(attempts)
            self.logger.warning(f'will reconnect in {delay}s')
            if hasattr(self.pubnub, 'event_loop'):
                self.run_async(self.delayed_reconnect_async(delay, attempts))

    async def delayed_reconnect_async(self, delay, attempt):
        self.stop_event = self.get_new_stop_event()
        await asyncio.sleep(delay)

        request = Subscribe(self.pubnub).timetoken(self.get_timetoken()).cancellation_event(self.stop_event)

        if self.invocation.channels:
            request.channels(self.invocation.channels)
        if self.invocation.groups:
            request.channel_groups(self.invocation.groups)

        if self.invocation.region:
            request.region(self.invocation.region)

        if feature_enabled('PN_MAINTAIN_PRESENCE_STATE') and hasattr(self.pubnub, 'state_container'):
            state_container = self.pubnub.state_container
            request.state(state_container.get_state(self.invocation.channels))

        response = await request.future()

        if isinstance(response, PubNubException):
            self.logger.warning(f'Reconnect failed: {str(response)}')
            self.failure(str(response), attempt, self.get_timetoken())

        elif response.status.error:
            self.logger.warning(f'Reconnect failed: {response.status.error_data.__dict__}')
            self.failure(response.status.error_data, attempt, self.get_timetoken())
        else:
            cursor = response.result['t']
            timetoken = int(self.invocation.timetoken) if self.invocation.timetoken else cursor['t']
            region = cursor['r']
            messages = response.result['m']
            self.success(timetoken=timetoken, region=region, messages=messages)

    def stop(self):
        self.logger.debug(f'stop called on {self.__class__.__name__}')
        if self.stop_event:
            self.logger.debug(f'stop_event({id(self.stop_event)}).set() called on {self.__class__.__name__}')
            self.stop_event.set()
            if self.task:
                try:
                    self.task.cancel()
                except asyncio.exceptions.CancelledError:
                    pass


class HandshakeReconnectEffect(ReconnectEffect):
    def give_up(self, reason: PubNubException, attempt: int, timetoken: int = 0):
        self.event_engine.trigger(
            events.HandshakeReconnectGiveupEvent(reason, attempt, timetoken)
        )

    def failure(self, reason: PubNubException, attempt: int, timetoken: int = 0):
        self.event_engine.trigger(
            events.HandshakeReconnectFailureEvent(reason, attempt, timetoken)
        )

    def success(self, timetoken: str, region: Optional[int] = None, **kwargs):
        self.event_engine.trigger(
            events.HandshakeReconnectSuccessEvent(timetoken, region)
        )

    def get_timetoken(self):
        return 0


class ReceiveReconnectEffect(ReconnectEffect):
    def give_up(self, reason: PubNubException, attempt: int, timetoken: int = 0):
        self.event_engine.trigger(
            events.ReceiveReconnectGiveupEvent(reason, attempt, timetoken)
        )

    def failure(self, reason: PubNubException, attempt: int, timetoken: int = 0):
        self.event_engine.trigger(
            events.ReceiveReconnectFailureEvent(reason, attempt, timetoken)
        )

    def success(self, timetoken: str, region: Optional[int] = None, messages=None):

        self.event_engine.trigger(
            events.ReceiveReconnectSuccessEvent(timetoken=timetoken, region=region, messages=messages)
        )

    def get_timetoken(self):
        return int(self.invocation.timetoken)


class HeartbeatEffect(Effect):
    def run(self):
        channels = self.invocation.channels
        groups = self.invocation.groups
        if hasattr(self.pubnub, 'event_loop'):
            self.stop_event = self.get_new_stop_event()
            self.run_async(self.heartbeat(channels=channels, groups=groups, stop_event=self.stop_event))

    async def heartbeat(self, channels, groups, stop_event):
        request = Heartbeat(self.pubnub).channels(channels).channel_groups(groups).cancellation_event(stop_event)

        if feature_enabled('PN_MAINTAIN_PRESENCE_STATE') and hasattr(self.pubnub, 'state_container'):
            state_container = self.pubnub.state_container
            request.state(state_container.get_state(self.invocation.channels))

        response = await request.future()

        if isinstance(response, PubNubException):
            self.logger.warning(f'Heartbeat failed: {str(response)}')
            self.event_engine.trigger(events.HeartbeatFailureEvent(channels=channels, groups=groups,
                                                                   reason=response.status.error_data, attempt=1))
        elif response.status.error:
            self.logger.warning(f'Heartbeat failed: {response.status.error_data.__dict__}')
            self.event_engine.trigger(events.HeartbeatFailureEvent(channels=channels, groups=groups,
                                                                   reason=response.status.error_data, attempt=1))
        else:
            self.event_engine.trigger(events.HeartbeatSuccessEvent(channels=channels, groups=groups))


class HeartbeatWaitEffect(Effect):
    def __init__(self, pubnub_instance, event_engine_instance, invocation: invocations.HeartbeatWaitInvocation) -> None:
        super().__init__(pubnub_instance, event_engine_instance, invocation)
        self.heartbeat_interval = pubnub_instance.config.heartbeat_interval

    def run(self):
        if hasattr(self.pubnub, 'event_loop'):
            self.stop_event = self.get_new_stop_event()
            self.run_async(self.heartbeat_wait(self.heartbeat_interval, stop_event=self.stop_event))

    async def heartbeat_wait(self, wait_time: int, stop_event):
        try:
            await asyncio.sleep(wait_time)
            self.event_engine.trigger(events.HeartbeatTimesUpEvent())
        except asyncio.CancelledError:
            pass


class HeartbeatLeaveEffect(Effect):
    def run(self):
        channels = self.invocation.channels
        groups = self.invocation.groups
        if hasattr(self.pubnub, 'event_loop'):
            self.stop_event = self.get_new_stop_event()
            self.run_async(self.leave(channels=channels, groups=groups, stop_event=self.stop_event))

    async def leave(self, channels, groups, stop_event):
        leave_request = Leave(self.pubnub).channels(channels).channel_groups(groups).cancellation_event(stop_event)
        leave = await leave_request.future()

        if leave.status.error:
            self.logger.warning(f'Heartbeat failed: {leave.status.error_data.__dict__}')


class HeartbeatDelayedEffect(Effect):
    def __init__(self, pubnub_instance, event_engine_instance,
                 invocation: Union[invocations.PNManageableInvocation, invocations.PNCancelInvocation]) -> None:
        super().__init__(pubnub_instance, event_engine_instance, invocation)
        self.reconnection_policy = pubnub_instance.config.reconnect_policy
        self.max_retry_attempts = pubnub_instance.config.maximum_reconnection_retries
        self.interval = pubnub_instance.config.RECONNECTION_INTERVAL
        self.min_backoff = pubnub_instance.config.RECONNECTION_MIN_EXPONENTIAL_BACKOFF
        self.max_backoff = pubnub_instance.config.RECONNECTION_MAX_EXPONENTIAL_BACKOFF

    def run(self):
        if self.reconnection_policy is PNReconnectionPolicy.NONE or self.invocation.attempts > self.max_retry_attempts:
            self.event_engine.trigger(events.HeartbeatGiveUpEvent(channels=self.invocation.channels,
                                                                  groups=self.invocation.groups,
                                                                  reason=self.invocation.reason,
                                                                  attempt=self.invocation.attempts))

        if hasattr(self.pubnub, 'event_loop'):
            self.stop_event = self.get_new_stop_event()
            self.run_async(self.heartbeat(channels=self.invocation.channels, groups=self.invocation.groups,
                                          attempt=self.invocation.attempts, stop_event=self.stop_event))

    async def heartbeat(self, channels, groups, attempt, stop_event):
        if self.reconnection_policy is PNReconnectionPolicy.NONE or self.invocation.attempts > self.max_retry_attempts:
            self.event_engine.trigger(events.HeartbeatGiveUpEvent(channels=self.invocation.channels,
                                                                  groups=self.invocation.groups,
                                                                  reason=self.invocation.reason,
                                                                  attempt=self.invocation.attempts))
        request = Heartbeat(self.pubnub).channels(channels).channel_groups(groups).cancellation_event(stop_event)
        delay = self.calculate_reconnection_delay(attempt)
        self.logger.warning(f'Will retry to Heartbeat in {delay}s')
        await asyncio.sleep(delay)

        response = await request.future()
        if isinstance(response, PubNubException):
            self.logger.warning(f'Heartbeat failed: {str(response)}')
            self.event_engine.trigger(events.HeartbeatFailureEvent(channels=channels, groups=groups,
                                                                   reason=response.status.error_data,
                                                                   attempt=attempt))
        elif response.status.error:
            self.logger.warning(f'Heartbeat failed: {response.status.error_data.__dict__}')
            self.event_engine.trigger(events.HeartbeatFailureEvent(channels=channels, groups=groups,
                                                                   reason=response.status.error_data,
                                                                   attempt=attempt))
        else:
            self.event_engine.trigger(events.HeartbeatSuccessEvent(channels=channels, groups=groups))


class EffectFactory:
    _managed_invocations = {
        invocations.HandshakeInvocation.__name__: HandshakeEffect,
        invocations.ReceiveMessagesInvocation.__name__: ReceiveMessagesEffect,
        invocations.HandshakeReconnectInvocation.__name__: HandshakeReconnectEffect,
        invocations.ReceiveReconnectInvocation.__name__: ReceiveReconnectEffect,
        invocations.HeartbeatInvocation.__name__: HeartbeatEffect,
        invocations.HeartbeatWaitInvocation.__name__: HeartbeatWaitEffect,
        invocations.HeartbeatDelayedHeartbeatInvocation.__name__: HeartbeatDelayedEffect,
        invocations.HeartbeatLeaveInvocation.__name__: HeartbeatLeaveEffect,
    }

    def __init__(self, pubnub_instance, event_engine_instance) -> None:
        self._pubnub = pubnub_instance
        self._event_engine = event_engine_instance

    def create(self, invocation: invocations.PNInvocation) -> Effect:
        if invocation.__class__.__name__ not in self._managed_invocations:
            raise PubNubException(errormsg=f"Unhandled Invocation: {invocation.__class__.__name__}")
        return self._managed_invocations[invocation.__class__.__name__](self._pubnub, self._event_engine, invocation)


class EmitEffect:
    pubnub: PubNub
    message_worker: BaseMessageWorker

    def set_pn(self, pubnub: PubNub):
        self.pubnub = pubnub
        self.message_worker = BaseMessageWorker(pubnub)

    def emit(self, invocation: invocations.PNEmittableInvocation):
        if isinstance(invocation, invocations.EmitMessagesInvocation):
            self.emit_message(invocation)
        if isinstance(invocation, invocations.EmitStatusInvocation):
            self.emit_status(invocation)

    def emit_message(self, invocation: invocations.EmitMessagesInvocation):
        self.message_worker._listener_manager = self.pubnub._subscription_manager._listener_manager
        for message in invocation.messages:
            subscribe_message = SubscribeMessage().from_json(message)
            self.message_worker._process_incoming_payload(subscribe_message)

    def emit_status(self, invocation: invocations.EmitStatusInvocation):
        pn_status = PNStatus()
        pn_status.category = invocation.status
        pn_status.error = False
        self.pubnub._subscription_manager._listener_manager.announce_status(pn_status)



from pubnub.event_engine.models import invocations
from pubnub.event_engine import effects


class Dispatcher:
    _pubnub = None
    _effects_factory = None

    def __init__(self, event_engine) -> None:
        self._event_engine = event_engine
        self._managed_effects = {}
        self._effect_emitter = effects.EmitEffect()

    def set_pn(self, pubnub_instance):
        self._pubnub = pubnub_instance
        self._effect_emitter.set_pn(pubnub_instance)

    def dispatch_effect(self, invocation: invocations.PNInvocation):
        if not self._effects_factory:
            self._effects_factory = effects.EffectFactory(self._pubnub, self._event_engine)

        if isinstance(invocation, invocations.PNEmittableInvocation):
            self.emit_effect(invocation)

        elif isinstance(invocation, invocations.PNManageableInvocation):
            self.dispatch_managed_effect(invocation)

        elif isinstance(invocation, invocations.PNCancelInvocation):
            self.dispatch_cancel_effect(invocation)

    def emit_effect(self, effect: invocations.PNInvocation):
        self._effect_emitter.emit(effect)

    def dispatch_managed_effect(self, invocation: invocations.PNInvocation):
        effect = self._effects_factory.create(invocation)
        effect.run()
        self._managed_effects[invocation.__class__.__name__] = effect

    def dispatch_cancel_effect(self, invocation: invocations.PNInvocation):
        if invocation.cancel_effect in self._managed_effects:
            self._managed_effects[invocation.cancel_effect].stop()
            del self._managed_effects[invocation.cancel_effect]
